//
// Cforall Version 1.0.0 Copyright (C) 2015 University of Waterloo
//
// The contents of this file are covered under the licence agreement in the
// file "LICENCE" distributed with Cforall.
//
// RemoveInit.cc --
//
// Author           : Rob Schluntz
// Created On       : Mon May 18 07:44:20 2015
// Last Modified By : Rob Schluntz
// Last Modified On : Mon Jan 11 14:41:26 2016
// Update Count     : 118
//

#include <stack>
#include <list>
#include "RemoveInit.h"
#include "SynTree/Declaration.h"
#include "SynTree/Type.h"
#include "SynTree/Expression.h"
#include "SynTree/Statement.h"
#include "SynTree/Initializer.h"
#include "SynTree/Mutator.h"
#include "GenPoly/PolyMutator.h"

namespace InitTweak {
	namespace {
		const std::list<Label> noLabels;
	}

	class RemoveInit : public GenPoly::PolyMutator {
	  public:
		RemoveInit();
		virtual ObjectDecl * mutate( ObjectDecl *objDecl );
		virtual DeclarationWithType * mutate( FunctionDecl *functionDecl );

		virtual Statement * mutate( ReturnStmt * returnStmt );

	  protected:
		std::list<DeclarationWithType*> returnVals;
		UniqueName tempNamer;
		std::string funcName;
	};

	class CtorDtor : public GenPoly::PolyMutator {
	  public:
		// CtorDtor();

		virtual ObjectDecl * mutate( ObjectDecl * );

		virtual CompoundStmt * mutate( CompoundStmt * compoundStmt );

	  protected:
		typedef std::map< std::string, DeclarationWithType * > MMMMAP;
		std::stack< MMMMAP > constructedObjects;

		// to be added before block ends - use push_front so order is correct
		std::list< Statement * > destructorStmts;
	};

	void tweak( std::list< Declaration * > translationUnit ) {
		RemoveInit remover;
		CtorDtor ctordtor;
		mutateAll( translationUnit, remover );
		mutateAll( translationUnit, ctordtor );
	}

	RemoveInit::RemoveInit() : tempNamer( "_retVal" ) {}

	// in the case where an object has an initializer and a polymorphic type, insert an assignment immediately after the
	// declaration. This will (seemingly) cause the later phases to do the right thing with the assignment
	ObjectDecl *RemoveInit::mutate( ObjectDecl *objDecl ) {
		if (objDecl->get_init() && dynamic_cast<TypeInstType*>(objDecl->get_type())) {
			if (SingleInit * single = dynamic_cast<SingleInit*>(objDecl->get_init())) {
				UntypedExpr *assign = new UntypedExpr( new NameExpr( "?=?" ) );
				assign->get_args().push_back( new AddressExpr (new NameExpr( objDecl->get_name() ) ) );
				assign->get_args().push_back( single->get_value()->clone() );
				stmtsToAddAfter.push_back(new ExprStmt(noLabels, assign));
			} // if
		} // if
		return objDecl;
	}

	Statement *RemoveInit::mutate( ReturnStmt *returnStmt ) {
		// update for multiple return values
		assert( returnVals.size() == 0 || returnVals.size() == 1 );
		// hands off if the function returns an lvalue - we don't want to allocate a temporary if a variable's address
		// is being returned
		if ( returnStmt->get_expr() && returnVals.size() == 1 && funcName != "?=?" && ! returnVals.front()->get_type()->get_isLvalue()  ) {
			ObjectDecl *newObj = new ObjectDecl( tempNamer.newName(), DeclarationNode::NoStorageClass, LinkageSpec::C, 0, returnVals.front()->get_type()->clone(), 0 );
			stmtsToAdd.push_back( new DeclStmt( noLabels, newObj ) );

			UntypedExpr *assign = new UntypedExpr( new NameExpr( "?=?" ) );
			assign->get_args().push_back( new AddressExpr (new NameExpr( newObj->get_name() ) ) );
			assign->get_args().push_back( returnStmt->get_expr() );
			stmtsToAdd.push_back(new ExprStmt(noLabels, assign));

			returnStmt->set_expr( new VariableExpr( newObj ) );
		} // if
		return returnStmt;
	}

	DeclarationWithType* RemoveInit::mutate( FunctionDecl *functionDecl ) {
		std::list<DeclarationWithType*> oldReturnVals = returnVals;
		std::string oldFuncName = funcName;

		FunctionType * type = functionDecl->get_functionType();
		returnVals = type->get_returnVals();
		funcName = functionDecl->get_name();
		DeclarationWithType * decl = Mutator::mutate( functionDecl );
		returnVals = oldReturnVals;
		funcName = oldFuncName;
		return decl;
	}

	bool tryConstruct( ObjectDecl * objDecl ) {
		// xxx - handle designations
		return objDecl->get_init() == NULL ||
			( objDecl->get_init() != NULL && objDecl->get_init()->get_maybeConstructed() );
	}

	ExprStmt * makeCtorDtorStmt( std::string name, ObjectDecl * objDecl, std::list< Expression * > args ) {
		UntypedExpr * expr = new UntypedExpr( new NameExpr( name ) );
		expr->get_args().push_back( new VariableExpr( objDecl ) );
		expr->get_args().splice( expr->get_args().end(), args );
		return new ExprStmt( noLabels, expr );
	}

	// InitExpander ctor/dtor being marked as weak symbols
	// this is causing a bug - Rodolfo's InitExpander is being constructed and destructed
	// with different fields, which causes a segfault

	class InitExpander : public Visitor {
	  public:
	  InitExpander() {}
	  // ~InitExpander() {}
		virtual void visit( SingleInit * singleInit );
		virtual void visit( ListInit * listInit );
		std::list< Expression * > argList;
	};

	void InitExpander::visit( SingleInit * singleInit ) {
		argList.push_back( singleInit->get_value()->clone() );
	}

	void InitExpander::visit( ListInit * listInit ) {
		// xxx - for now, assume no nested list inits
		std::list<Initializer*>::iterator it = listInit->begin_initializers();
		for ( ; it != listInit->end_initializers(); ++it ) {
			(*it)->accept( *this );
		}
	}

	std::list< Expression * > makeInitList( Initializer * init ) {
		if ( init ) {
			InitExpander expander;
			// init->accept( expander );
			// std::list< Expression * > l = expander.argList;
			std::list< Expression * > l;
			return l;
		} else {
			std::list< Expression * > l;
			return l;
		}
	}

	ObjectDecl * CtorDtor::mutate( ObjectDecl * objDecl ) {
		// hands off if designated or if @=
		if ( tryConstruct( objDecl ) ) {
			ExprStmt * ctor = makeCtorDtorStmt( "?{}", objDecl, makeInitList( objDecl->get_init() ) );
			ExprStmt * dtor = makeCtorDtorStmt( "^?{}", objDecl, std::list< Expression * >() );

			// set_ctor...
			// need to remember init expression, in case no ctors exist
			// if ctor does exist, want to use ctor stmt instead of init
			objDecl->set_ctor( ctor );
			destructorStmts.push_front( dtor );
		}
		return objDecl;
	}

	CompoundStmt * CtorDtor::mutate( CompoundStmt * compoundStmt ) {
		CompoundStmt * ret = PolyMutator::mutate( compoundStmt );
		std::list< Statement * > &statements = ret->get_kids();
		if ( ! destructorStmts.empty() ) {
			statements.splice( statements.end(), destructorStmts );
		} // if
		return ret;
	}



} // namespace InitTweak

// Local Variables: //
// tab-width: 4 //
// mode: c++ //
// compile-command: "make install" //
// End: //
